/* Cria alguns mapas de normais para testes. */
/* Last edited on 2005-08-14 23:21:18 by stolfi */ 

#define PROG_NAME "cria_testes"
#define PROG_HELP \
  PROG_NAME 

#define _GNU_SOURCE

#include <imagem_vetores.h>
#include <imagem_valores.h>

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

/* PROTTIPOS DE FUNES */

void gera_imagem_teste
  ( char *nome, 
    void func (double X, double Y, double *Z, double *dZdX, double *dZdY),
    imagem_valores_t *IZ, 
    imagem_vetores_t *IN
  );
  /* Gera uma imagem de alturas a partir da funo {func}. Armazena as
    alturas (calculadas nos vrtices da grade) em {IZ}, e as normais
    mdias (calculadas no centro de cada pixel) em {IN}. A seguir
    escreve as duas imagens com nomes "{nome}-alturas.ival" e
    "{nome}-normais.ivet", respectivamente.

    A funo {func} recebe coordenadas {X,Y} dos pontos na imagem,
    convertidas para uma escala tal que o quadrado 
    {[-1 _ +1][-1 _ +1]} cabe justo na imagem, centrado.
    Ela deve devolver as alturas no intervalo {[-1 _ +1]} tambm. */
  
void func_teste_1(double X, double Y, double *Z, double *dZdX, double *dZdY);
void func_teste_2(double X, double Y, double *Z, double *dZdX, double *dZdY);
void func_teste_3(double X, double Y, double *Z, double *dZdX, double *dZdY);

int main(int argc,char** argv);

/* IMPLEMENTAES */

int main(int argc,char** argv)
  {
    if(argc != 1)
      { fprintf(stderr, "Chamada Correta: %s\n", PROG_HELP); return 1; }

    int ty = 24, tx = 32;
    
    imagem_vetores_t *IN = cria_imagem_vetores(tx, ty);
    imagem_valores_t *IZ = cria_imagem_valores(tx+1, ty+1);
    
    gera_imagem_teste("teste-1", func_teste_1, IZ, IN);
    gera_imagem_teste("teste-2", func_teste_2, IZ, IN);
    gera_imagem_teste("teste-3", func_teste_3, IZ, IN);

    return 0;
  }

void gera_imagem_teste
  ( char *nome, 
    void func(double X, double Y, double *Z, double *dZdX, double *dZdY),
    imagem_valores_t *IZ, 
    imagem_vetores_t *IN
  )
  {
    int tx = IN->tx; assert(IZ->tx == tx+1); double cx = 0.5*tx;
    int ty = IN->ty; assert(IZ->ty == ty+1); double cy = 0.5*ty;
    int t = (tx < ty ? tx : ty); /* Menor dimenso. */
    double scale = t/2.0;
    int NS = 6; /* Usa {NS*NS} amostras em cada pixel para derivadas. */
    
    int x, y;
    for (y = 0; y <= ty; y++)
      {
        for (x = 0; x <= tx; x++)
          {
            /* Converte {(x,y)} para {(X,Y)} em [-1.0 _ +1.0]: */
            double X = (x - cx)/scale;
            double Y = (y - cy)/scale;
            
            /* Calcula altura no vrtice {[y][x]}, ignora derivadas: */
            double Z, dZdX, dZdY;
            func(X, Y, &Z, &dZdX, &dZdY);
            IZ->pixel[y][x] = Z*scale;
            
            if ((y < ty) && (x < tx))
              { 
                vetor_t nrm;
                /* Calcula derivadas mdias dentro do pixel {[y][x]}: */
                Z = 0; dZdX = 0; dZdY = 0;
                int xs, ys;
                for (ys = 0; ys < NS; ys++)
                  for (xs = 0; xs < NS; xs++)
                    { /* Gera um ponto de amostra {(Xk,Yk)} dentro do pixel {[y][x]}: */
                      double Xk = X + (xs + 0.5)/(NS + 1)/scale;
                      double Yk = Y + (ys + 0.5)/(NS + 1)/scale;
                      /* Calcula funo e derivadas no ponto de amostra: */
                      double Zk, dZkdX, dZkdY;
                      func(Xk, Yk, &Zk, &dZkdX, &dZkdY);
                      /* Acumula funo e derivadas: */
                      Z += Zk; dZdX += dZkdX, dZdY += dZkdY;
                    }
                /* Calcula funo e derivadas mdias: */
                Z /= NS*NS; dZdX /= NS*NS; dZdY /= NS*NS;
                /* Muda escala de {X,Y,Z} para pixels: */
                Z = Z*scale;  /* Note que {dZdX} e {dZdY} no mudam. */
                /* Converte {dZdX,dZdY} para vetor normal, guarda: */
                double m = sqrt(1.0 + dZdX*dZdX + dZdY*dZdY);
                nrm.x = -dZdX/m;
                nrm.y = -dZdY/m;
                nrm.z = 1.0/m;
                IN->pixel[y][x] = nrm;
              }
          }
      }
    
    /* Escreve imagens: */
    char *fileName;
    
    fileName = NULL; 
    asprintf(&fileName, "%s-alturas.ival", nome);
    escreve_imagem_valores(fileName, IZ);
    free(fileName);
    
    fileName = NULL; 
    asprintf(&fileName, "%s-normais.ivet", nome);
    escreve_imagem_vetores(fileName, IN);
    free(fileName);
  }

void func_teste_1(double X, double Y, double *Z, double *dZdX, double *dZdY)
  {
    /* Rampa linear na diagonal: */
    double CX = 1/3.0, CY = 1/2.0;
    (*Z) = CX*X + CY*Y;
    (*dZdX) = CX;
    (*dZdY) = CY;
  }

void func_teste_2(double X, double Y, double *Z, double *dZdX, double *dZdY)
  {
    /* Lombada parablica na diagonal: */
    double CX = 1/3.0, CY = 1/2.0;
    double S = CX*X + CY*Y;
    (*Z) = 1-(S*S);
    (*dZdX) = -2.0*S*CX;
    (*dZdY) = -2.0*S*CY;
  }

void func_teste_3(double X, double Y, double *Z, double *dZdX, double *dZdY)
  {
    /* Calota hemisfrica: */
    double Zmin = 0.1; /* Nvel do fundo. */
    double R = 0.8, R2 = R*R; /* Raio da esfera. */
    (*Z) = Zmin; (*dZdX) = 0; (*dZdY) = 0; /* Fundo. */
    double D2 = X*X + Y*Y;
    if (D2 < R2) 
      { double S = R2 - D2;
        double dSdX = -2*X;
        double dSdY = -2*Y;
        double F = sqrt(S);
        if (F > Zmin)
          { double dFdS = 0.5/F;
            (*Z) = F;
            (*dZdX) = dFdS*dSdX;
            (*dZdY) = dFdS*dSdY;
          }
      }
  }
